import os
import shutil
import subprocess


def clone_repository(repo_url, repo_directory):
    subprocess_args = ["git", "clone", repo_url, repo_directory]
    try:
        subprocess.run(subprocess_args, check=True)
        print("\nTemplate cloned.")
    except subprocess.CalledProcessError:
        print("Failed.")
        exit(128)


def delete_git_folder(repo_directory):
    current_directory = os.getcwd()
    repo_path = os.path.join(current_directory, repo_directory)
    git_folder = os.path.join(repo_path, ".git")

    shutil.rmtree(git_folder)


def new_client(*args):
    if not args:
        print("You must specify the language for the client: cq23 new python my_bot")
        return

    language_map = {
        "raw": "https://github.com/CALED-Team/codequest23-raw-submission.git",
        "python": "https://github.com/CALED-Team/codequest23-python-submission.git",
        "cpp": "https://github.com/CALED-Team/codequest23-cpp-submission.git",
    }
    language = args[0]

    if language not in language_map.keys():
        return print(
            "Invalid language selected for the new bot. Language should be one of:",
            "[" + " - ".join(language_map.keys()) + "]",
        )

    if len(args) < 2:
        print(
            "You should specify a name for the new bot. A new folder with this name will be created so make sure"
            " there are no other folders with this name in the current directory. Example: cq23 new python my_bot"
        )
        return
    if len(args) > 2:
        print(
            "Unrecognised arguments were passed. Maybe you chose a name for your client that included spaces? If that"
            ' is the case make sure you wrap the name with double quotes. Example: cq23 new python "my bot"'
        )
        return

    repo_directory = args[1]

    clone_repository(language_map[language], repo_directory)
    delete_git_folder(repo_directory)
    print(
        f"New {language} client created in {repo_directory}. You can now go in the folder and run the game:"
    )
    print(f'> cd "{repo_directory}"')
    print("> cq23 run")



import os
import zipfile


def zip_current_directory():
    current_dir = os.getcwd()
    zip_filename = "submission.zip"
    excluded_files = [zip_filename, ".DS_Store", ".gitignore"]
    excluded_folders = [".game_files", ".idea", ".vscode", ".git"]

    with zipfile.ZipFile(zip_filename, "w", zipfile.ZIP_DEFLATED) as zipf:
        for root, dirs, files in os.walk(current_dir):

            # Remove bad files
            for excluded_file in excluded_files:
                if excluded_file in files:
                    files.remove(excluded_file)

            # Remove bad folders
            for excluded_folder in excluded_folders:
                if excluded_folder in dirs:
                    dirs.remove(excluded_folder)

            # Put all files in
            for file in files:
                file_path = os.path.join(root, file)
                zipf.write(file_path, arcname=os.path.relpath(file_path, current_dir))

            # Put all folders in
            for folder in dirs:
                folder_path = os.path.join(root, folder)
                arcname = os.path.relpath(folder_path, current_dir)
                zipf.write(folder_path, arcname=arcname)

    print(f"Submission file '{zip_filename}' created successfully!")


def zip(*args):
    zip_current_directory()



import os
import subprocess
import time

import boto3


def create_ec2_instance(
    volume_size,
    iam_role_arn,
    ami_id,
    sg_id,
    instance_name,
    instance_type="t2.small",
    key_name="aws5-ap-southeast-2",
):
    ec2 = boto3.client("ec2")

    # Create a new EC2 instance
    response = ec2.run_instances(
        ImageId=ami_id,
        InstanceType=instance_type,
        MinCount=1,
        MaxCount=1,
        BlockDeviceMappings=[
            {"DeviceName": "/dev/sda1", "Ebs": {"VolumeSize": volume_size}}
        ],
        IamInstanceProfile={"Arn": iam_role_arn},
        TagSpecifications=[
            {
                "ResourceType": "instance",
                "Tags": [{"Key": "Name", "Value": instance_name}],
            }
        ],
        KeyName=key_name,
        SecurityGroupIds=[sg_id],
    )

    # Extract the instance ID and IP address
    instance_id = response["Instances"][0]["InstanceId"]

    # Describe the instance to get the public IP address
    time.sleep(1)
    describe_response = ec2.describe_instances(InstanceIds=[instance_id])
    instance_ip = describe_response["Reservations"][0]["Instances"][0].get(
        "PublicIpAddress"
    )

    print("-------------------------------------------")
    print(f"Created EC2 instance with ID: {instance_id}")
    print(f"Public IP address: {instance_ip}")

    return instance_id, instance_ip


def check_ssh_readiness(instance_id):
    ec2 = boto3.client("ec2")

    while True:
        response = ec2.describe_instance_status(
            InstanceIds=[instance_id], IncludeAllInstances=True
        )
        instance_status = response["InstanceStatuses"]

        if not instance_status:
            print(f"No instance found with ID: {instance_id}")
            return False

        state = instance_status[0]["InstanceState"]["Name"]
        status = instance_status[0]["InstanceStatus"]["Status"]

        if state == "running" and status == "ok":
            print(
                f"Instance {instance_id} is running and ready to accept SSH connections."
            )
            return True
        elif state == "terminated":
            print(f"Instance {instance_id} is terminated.")
            return False

        print(
            f"Instance {instance_id} is not yet ready. Current state: {state}. Status: {status}"
        )
        print("Waiting...")
        time.sleep(10)


def terminate_instances_by_name(name):
    ec2 = boto3.client("ec2")

    response = ec2.describe_instances(
        Filters=[
            {"Name": "tag:Name", "Values": [name]},
            {"Name": "instance-state-name", "Values": ["running"]},
        ]
    )

    instance_ids = []
    reservations = response["Reservations"]

    if not reservations:
        print(f"No instances found with the name: {name}")
        return

    for reservation in reservations:
        for instance in reservation["Instances"]:
            instance_id = instance["InstanceId"]
            instance_ids.append(instance_id)

    if not instance_ids:
        print(f"No instances found with the name: {name}")
        return

    response = ec2.terminate_instances(InstanceIds=instance_ids)

    if "TerminatingInstances" in response:
        print("Terminating the following instances:")
        for instance in response["TerminatingInstances"]:
            print(f"- {instance['InstanceId']}")
    else:
        print("No instances terminated.")


def create_cq_instances(instance_type, *args):
    assert instance_type in ["worker", "builder"]

    ami_id = "ami-05f998315cca9bfe3"
    role_arn = os.getenv("CQ23CLI_IAM_ROLE_ARN")
    if not role_arn:
        return print("Role ARN is needed for creating new instances.")
    sg_id = os.getenv("CQ23CLI_SG_ID")
    if not sg_id:
        return print("SG ID is needed for creating new instances.")
    if not args or not str(args[0]).isnumeric():
        return print("Invalid number of instances:", args[0])

    instance_count = int(args[0])
    instance_ids = []
    instance_ips = []
    for i in range(instance_count):
        i_id, i_ip = create_ec2_instance(
            20, role_arn, ami_id, sg_id, f"cq-{instance_type}", "t2.small"
        )
        instance_ids.append(i_id)
        instance_ips.append(i_ip)

    time.sleep(5)

    for i in range(len(instance_ips)):
        if not check_ssh_readiness(instance_ids[i]):
            print("Failed!")
            continue
        subprocess.run(["make", "deploy", f"ip={instance_ips[i]}"], check=True)

    print(f"All {instance_type}s built and deployed.")
    print("\n".join(instance_ips))



import json
import os
import subprocess
import sys
import webbrowser
from multiprocessing import Process

from web_server import flask_api

from . import docker_tools

MATCH_TIMEOUT_SECONDS = 12 * 60  # This should ideally match the one in game worker


def run_gui():
    replay_files_directory = os.path.join(os.getcwd(), "replay_files")
    # gui_directory = os.path.join(os.getcwd(), "gui")
    gui_process = Process(target=flask_api.start, args=(replay_files_directory,))
    gui_process.start()
    # webbrowser.open("file://" + os.path.join(gui_directory, "index.html"))
    webbrowser.open("https://watch.codequest.club/?base_url=http://127.0.0.1:2023/")
    return gui_process


def stop_gui(gui_process):
    func = os.environ.get("werkzeug.server.shutdown")
    if func is None:
        print("Can't stop gracefully, killing the server process.")
        gui_process.terminate()
    else:
        func()
        gui_process.join(timeout=5)
        if gui_process.is_alive():
            gui_process.terminate()


def run_gcs(gcs_folder_name, game_map=None):
    gcs_src_dir = os.path.join(gcs_folder_name, "src")
    clients_file_content = [
        {
            "id": "1",
            "name": "Your Code - 1",
            "image": docker_tools.get_client_image_tag(),
        },
        {
            "id": "2",
            "name": "Your Code - 2",
            "image": docker_tools.get_client_image_tag(),
        },
    ]
    clients_file_address = "clients.json"

    with open(os.path.join(gcs_src_dir, clients_file_address), "w") as f:
        f.write(json.dumps(clients_file_content))

    subprocess_args = [
        sys.executable,
        "controller.py",
        docker_tools.get_server_image_tag(),
        clients_file_address,
    ]

    if game_map:
        subprocess_args.append("--server-arg")
        subprocess_args.append("-m " + str(game_map))

    print("Starting the game interface...")
    gui_process = run_gui()

    print("Starting the game...")
    subprocess.run(subprocess_args, timeout=MATCH_TIMEOUT_SECONDS, cwd=gcs_src_dir)
    print("Game finished.")

    print("Stopping the game interface...")
    stop_gui(gui_process)

    os.remove(os.path.join(gcs_src_dir, clients_file_address))



from main.utils import admin

from .aws import create_cq_instances, terminate_instances_by_name


def new_workers(*args):
    create_cq_instances("worker", *args)


def destroy_workers(*args):
    if (
        input("Are you sure you want to destroy all workers? (y)").lower().strip()
        != "y"
    ):
        return print("Cancelled.")
    terminate_instances_by_name("cq-worker")


@admin
def command(*args):
    if not args or args[0] not in ["new", "destroy"]:
        print("new or destroy")

    if args[0] == "new":
        new_workers(*args[1:])
    elif args[0] == "destroy":
        destroy_workers(*args[1:])



import os.path

from flask import Flask, jsonify, request, send_file
from flask_cors import CORS

app = Flask(__name__)
CORS(app)
ROOT_DIRECTORY = None
PORT = None


def get_full_path_or_404(file_name):
    full_file_path = os.path.join(ROOT_DIRECTORY, file_name)
    print(full_file_path)
    if not os.path.exists(full_file_path):
        return False, (jsonify({"message": "Replay file not found."}), 404)

    return True, full_file_path


@app.route("/", methods=["GET"])
def heartbeat():
    return "I'm up"


@app.route("/download_file", methods=["GET"])
def download_file():
    assert ROOT_DIRECTORY
    file_name = request.args["file_name"]
    valid, response = get_full_path_or_404(file_name)
    if not valid:
        return response
    return send_file(response)


@app.route("/get_replay_file_url", methods=["GET"])
def get_file_url():
    assert ROOT_DIRECTORY and PORT
    file_name = request.args["file_name"]

    valid, response = get_full_path_or_404(file_name)
    if not valid:
        return response

    return jsonify(
        {
            "message": "ok",
            "url": f"http://127.0.0.1:{PORT}/download_file?file_name={file_name}",
        }
    )


@app.route("/get_replay_file_content", methods=["GET"])
def get_file_contents():
    assert ROOT_DIRECTORY
    file_name = request.args["file_name"]

    valid, response = get_full_path_or_404(file_name)
    if not valid:
        return response

    with open(response) as f:
        file_content = f.read()

    return jsonify({"message": "ok", "content": file_content})


def start(replay_files_directory, port=2023, debug=False):
    global ROOT_DIRECTORY, PORT
    ROOT_DIRECTORY = replay_files_directory
    PORT = port

    app.run(debug=debug, port=PORT, threaded=True)



import sys

from admin.builder import command as builder
from admin.worker import command as worker
from check.command import check
from cleanup.command import cleanup
from new_client.command import new_client
from replay.command import replay
from run_game.command import run_game
from zip.command import zip

from .utils import restore_cwd


def help_message():
    message = (
        "Available commands:\n\n"
        + "> cq23 new python my_bot\n"
        + "> cq23 run\n"
        + "> cq23 run map=<map name>\n"
        + "> cq23 replay\n"
        + "> cq23 zip\n"
        + "> cq23 check\n"
        + "> cq23 cleanup\n\n"
        + "If you need help with the competition, post a message in Discord or email us at info@codequest.club."
    )
    print(message)


@restore_cwd
def route_command():
    command_args = sys.argv[1:]

    first_arg_mapping = {
        "new": new_client,
        "run": run_game,
        "replay": replay,
        "cleanup": cleanup,
        "zip": zip,
        "check": check,
        "worker": worker,
        "builder": builder,
    }

    if not command_args or command_args[0].lower() not in first_arg_mapping.keys():
        help_message()
    else:
        first_arg_mapping[command_args[0].lower()](*command_args[1:])



import os
import shutil
import subprocess

import docker
from colorama import Fore
from colorama import init as color_init


def is_git_installed():
    try:
        subprocess.run(["git", "--version"], capture_output=True, check=True)
        return True
    except (FileNotFoundError, subprocess.CalledProcessError):
        return False


def is_docker_installed():
    try:
        client = docker.from_env()
        client.ping()
        return True
    except (docker.errors.DockerException, docker.errors.APIError):
        return False


def has_enough_disk_space(minimum_space_gb):
    root_directory = os.path.abspath(os.sep)
    total, used, free = shutil.disk_usage(root_directory)
    free_gb = free // (2**30)  # Convert bytes to GB
    return free_gb >= minimum_space_gb


def check(*args):
    color_init(autoreset=True)
    git_installed = is_git_installed()
    docker_installed = is_docker_installed()
    enough_disk_space = has_enough_disk_space(5)

    boolean_map = {True: Fore.GREEN + "Yes", False: Fore.RED + "No"}

    print(f"Git installed: {boolean_map[git_installed]}")
    print(f"Docker installed and running: {boolean_map[docker_installed]}")
    print(f"Enough disk space: {boolean_map[enough_disk_space]}")

    if git_installed and docker_installed and enough_disk_space:
        print(Fore.LIGHTGREEN_EX + "You are all set!")
    else:
        print(Fore.YELLOW + "Some requirements are not met.")
        if not docker_installed:
            print(
                Fore.YELLOW
                + "-> If you have installed Docker Desktop, you need to make sure it's running. Search "
                "for Docker in your applications and open it. That will start Docker on your computer."
            )



import os
import shutil
import subprocess

from . import docker_tools
from .gcs import run_gcs


def clone_or_pull_repository(repository_url, folder_path):
    if not os.path.exists(folder_path):
        # Clone the repository if the folder doesn't exist
        subprocess.run(["git", "clone", repository_url, folder_path])
    else:
        # Pull the latest changes if the folder already exists
        current_dir = os.getcwd()
        os.chdir(folder_path)
        subprocess.run(["git", "pull"])
        os.chdir(current_dir)
    print(folder_path + " cloned successfully.")


def extract_map_from_command_args(args):
    map_arg = list(filter(lambda x: str(x).startswith("map="), args))
    if not map_arg:
        return None
    if len(map_arg) > 1:
        raise Exception("Multiple `map` arguments provided.")

    # We can also validate the map name here if we want...

    return map_arg[0][4:]


def copy_container_logs(game_files_abs_path, gcs_folder_name):
    src_dir = os.path.join(os.path.join(gcs_folder_name, "src"), "container_logs")
    dst_dir = os.path.join(game_files_abs_path, "replay_files")
    for filename in os.listdir(src_dir):
        src_file = os.path.join(src_dir, filename)
        dst_file = os.path.join(dst_dir, filename)
        # Copy the file to the destination directory
        shutil.copy2(src_file, dst_file)


def run_game(*args):
    docker_tools.ensure_docker_client_exists()
    game_files_dir = ".game_files"
    gcs_folder_name = "gcs"
    # gui_folder_name = "gui"
    gcs_repo = "https://github.com/CALED-Team/game-communication-system.git"
    # gui_repo = "https://github.com/CALED-Team/game-gui-23.git"

    docker_tools.check_dockerfile_exists()
    docker_tools.build_and_tag_image(docker_tools.get_client_image_tag())

    if not os.path.exists(game_files_dir):
        os.makedirs(game_files_dir)
    os.chdir(game_files_dir)
    game_files_abs_path = os.getcwd()

    clone_or_pull_repository(gcs_repo, gcs_folder_name)
    # clone_or_pull_repository(gui_repo, gui_folder_name)
    docker_tools.pull_latest_game_server()
    run_gcs(gcs_folder_name, extract_map_from_command_args(args))
    docker_tools.copy_replay_files(game_files_abs_path)
    copy_container_logs(game_files_abs_path, gcs_folder_name)

    # Go back to the same directory as we were (one up)
    os.chdir(os.path.dirname(os.getcwd()))



import os
import time
import webbrowser
from multiprocessing import Process

from web_server import flask_api


def run_gui(replay_files_directory):
    gui_process = Process(target=flask_api.start, args=(replay_files_directory,))
    gui_process.start()
    webbrowser.open("https://watch.codequest.club/?base_url=localhost:2023/")
    return gui_process


def replay(*args):
    if not args:
        replays_folder = os.path.join(
            os.getcwd(), os.path.join(".game_files", "replay_files")
        )
        if not os.path.exists(replays_folder):
            print("No replay files available. You should first run a game: cq23 run")
            return
        process = run_gui(replays_folder)
        time.sleep(5 * 60)
        process.terminate()
    else:
        match_id = args[0]
        if not str(match_id).isnumeric():
            print("Match id should be a number: cq23 replay 123")
            return
        webbrowser.open(
            f"https://watch.codequest.club/?base_url=api.codequest.club/api/matches/{match_id}/"
        )
